/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.servicemix.components.xslt;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.StringWriter;
import java.net.URL;
import java.util.Iterator;
import java.util.Map;

import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Templates;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.servicemix.bean.support.TransformBeanSupport;
import org.apache.servicemix.jbi.jaxp.BytesSource;
import org.apache.servicemix.jbi.jaxp.SourceTransformer;
import org.apache.servicemix.jbi.jaxp.StringSource;
import org.springframework.core.io.Resource;
import org.w3c.dom.Document;
import org.w3c.dom.Node;

/**
 * An <a href="http://www.w3.org/TR/xslt">XSLT</a> based JBI component using <a
 * href="http://java.sun.com/xml/jaxp/">JAXP</a> to perform the XSLT
 * transformation.
 * 
 * @version $Revision: 464270 $
 */
public class XsltComponent extends TransformBeanSupport {
	
	private static final Log log = LogFactory.getLog(XsltComponent.class);
    private TransformerFactory transformerFactory;
    private Source xsltSource;
    private Resource xsltResource;
    private Templates templates;
    private boolean disableOutput;
    private boolean useStringBuffer = true;
    private boolean forceDocIfDom = true;
    private Map xsltParameters;

    /**
     * @return the forceDocIfDom
     */
    public boolean isForceDocIfDom() {
        return forceDocIfDom;
    }

    /**
     * @param forceDocIfDom the forceDocIfDom to set
     */
    public void setForceDocIfDom(boolean forceDocIfDom) {
        this.forceDocIfDom = forceDocIfDom;
    }

    // Properties
    // -------------------------------------------------------------------------
    public TransformerFactory getTransformerFactory() {
    	log.debug("XSLT Transformer::getTranformerFactory::->start");
        if (transformerFactory == null) {
        	log.debug("XSLT Transformer::TransformerFactory is null::create one");
            transformerFactory = TransformerFactory.newInstance();
            log.debug("XSLT Transformer::TransformerFactory created");
        }
        log.debug("XSLT Transformer::getTransformerFactory::transformerFactory " + ((transformerFactory != null) ? "transformerFactory is not null " : "transformerFactory IS NULL"));
        return transformerFactory;
    }

    public void setTransformerFactory(TransformerFactory transformerFactory) {
        this.transformerFactory = transformerFactory;
    }

    public Source getXsltSource() throws Exception {
    	log.debug("XSLT Transformer::getXsltSource::->start");
        if (xsltSource == null) {
        	log.debug("XSLT Transformer::getXsltSource::xsltSource is null create it ");
            // lets create a new one each time
            // as we can only read a stream once
            xsltSource = createXsltSource();
        	log.debug("XSLT Transformer::getXsltSource::xsltSource created ");
        }
    	log.debug("XSLT Transformer::getXsltSource::xsltSource " + ((xsltSource != null) ? "XSLT Source is not null " : "xsltSource IS NULL"));
        return xsltSource;
    }

    public void setXsltSource(Source xsltSource) {
        this.xsltSource = xsltSource;
    }

    public Resource getXsltResource() {
        return xsltResource;
    }

    public void setXsltResource(Resource xsltResource) {
        this.xsltResource = xsltResource;
    }

    public boolean isDisableOutput() {
        return disableOutput;
    }

    public void setDisableOutput(boolean disableOutput) {
        this.disableOutput = disableOutput;
    }

    public boolean isUseStringBuffer() {
        return useStringBuffer;
    }

    public void setUseStringBuffer(boolean useStringBuffer) {
        this.useStringBuffer = useStringBuffer;
    }

    /**
     * @return the xsltParameters
     */
    public Map getXsltParameters() {
        return xsltParameters;
    }

    /**
     * @param xsltParameters the xsltParameters to set
     */
    public void setXsltParameters(Map xsltParameters) {
        this.xsltParameters = xsltParameters;
    }

    // Implementation methods
    // -------------------------------------------------------------------------
    protected boolean transform(MessageExchange exchange, NormalizedMessage in, NormalizedMessage out)
            throws MessagingException {
        synchronized (this){
    		try {
    			log.debug("XSLT Transformer::transform::->start");
    			Transformer transformer = createTransformer(exchange, in);
    			log.debug("XSLT Transforme::Transformer Created");
    			configureTransformer(transformer, exchange, in);
    			log.debug("XSLT Transforme::Transformer Configured");
    			copyPropertiesAndAttachments(exchange, in, out);
    			log.debug("XSLT Transforme::Properties and attachment copied");
    			transformContent(transformer, exchange, in, out);
    			log.debug("XSLT Transforme::Content Transformed");
    			return shouldOutputResult(transformer);
            
    		}
        	catch (Exception e) {
        		throw new MessagingException("Failed to transform: " + e, e);
        	}
    	}
    }

    protected void transformContent(Transformer transformer, MessageExchange exchange, NormalizedMessage in, NormalizedMessage out) throws TransformerException, MessagingException, ParserConfigurationException {
        Source src = in.getContent();
        //System.out.println("*************** Source Class is "+src.getClass().getName());
        log.debug("XSLT Transformer::transformContent::->start");
        log.debug("XSLT Transformer::transformContent::->Source Class is "+src.getClass().getName());
        SourceTransformer st = new SourceTransformer();
        String str = st.toString(src);
        
        src = new StringSource(str);
        
        log.debug("XSLT Transformer::transformContent::->String source is " + str);
        if (src instanceof DOMSource) {
        	log.debug("XSLT Transformer::transformContent::->Src is DOMSOurce");
            Node n = ((DOMSource) src).getNode();
            if (n instanceof Document == false) {
                Document doc = DocumentBuilderFactory.newInstance().newDocumentBuilder().newDocument();
                doc.appendChild(doc.importNode(n, true));
                src = new DOMSource(doc);
            }
        }
        if (isUseStringBuffer()) {
        	log.debug("XSLT Transformer::transformContent::->Is Use StringBuffer " + isUseStringBuffer());
            StringWriter buffer = new StringWriter();
            Result result = new StreamResult(buffer);
            transformer.transform(src, result);
            out.setContent(new StringSource(buffer.toString()));
            log.debug("XSLT Transformer::transformContent::->OutMessage set");
        }
        else {
        	log.debug("XSLT Transformer::transformContent::-> Tasnaforming Content ");
            ByteArrayOutputStream buffer = new ByteArrayOutputStream();
            Result result = new StreamResult(buffer);
            transformer.transform(src, result);
            out.setContent(new BytesSource(buffer.toByteArray()));
            log.debug("XSLT Transformer::transformContent::-> OutMesageSet not use StringBuffer ");
        }
    }

    /**
     * Should we disable output of the result of the XSLT?
     */
    protected boolean shouldOutputResult(Transformer transformer) { 
    	log.debug("XSLT Transformer::shouldOutputResult::start");
    	if (disableOutput) {
    		log.debug("XSLT Transformer::shouldOutputResult::return false");
            return false;
        }
    	log.debug("XSLT Transformer::shouldOutputResult::return true");
        return true;
        /**
         * String value = transformer.getOutputProperty("disableOutput"); return
         * value == null || !value.equals("true");
         */
    }

    protected Source createXsltSource() throws Exception {
    	log.debug("XSLT Transformer::createXsltSource::start");
        if (xsltResource != null) {
        	log.debug("XSLT Transformer::xsltResource != null parse and retunrn it");
            return new DOMSource(parse(xsltResource));
        }
        log.debug("XSLT Transformer::xsltResource is null return null");
        return null;
    }

    protected Document parse(Resource res) throws Exception {
        URL url = null;
        try {
            res.getURL();
        } catch (IOException e) {
            log.error(e);
        }
        DocumentBuilder builder = new SourceTransformer().createDocumentBuilder();
        return builder.parse(res.getInputStream(), url != null ? url.toExternalForm() : null);
    }

    
    public Templates getTemplates() throws Exception {
    	log.debug("XSLT Transformer::getTemplates::start");
        if (templates == null) {
        	log.debug("XSLT Transformer::getTemplates::templates is NULL call createTemplates");
        	templates = createTemplates();
            log.debug("XSLT Transformer::getTemplates::templates created");
        }
        log.debug("XSLT Transformer::getTemplates::returning templates");
        log.debug("XSLT Transformer::getTemplates::templates "+ ( (templates != null) ? " Templastes is not null" : "Templates is NULL"));
        return templates;
    }

    /**
     * Factory method to create a new transformer instance
     */
    protected Templates createTemplates() throws Exception {
    	log.debug("XSLT Transformer::createTemplates::start::getting XSLTSource");
        Source source = getXsltSource();
        log.debug("XSLT Transformer::createTemplates::Source getted" + ( (source != null) ? "XsltSource Is not null" : "XsltSource is NULL"));
      
        Templates t = getTransformerFactory().newTemplates(source);
        log.debug("XSLT Transformer::createTemplates::templates created" + ( (t != null) ? "templates Is not null" : "Templates is NULL"));
        return t;
    }

    /**
     * Factory method to create a new transformer instance
     */
    protected Transformer createTransformer(MessageExchange exchange, NormalizedMessage in)
            throws Exception {
    	log.debug("XSLT Transformer::createTransformer::start");
    	log.debug("XSLT Transformer::createTransformer::getting source");
    	Source source = getXsltSource();
        
        if (source == null) {
        	log.debug("XSLT Transformer::createTransformer::source IS NULL");
        	log.debug("XSLT Transformer::createTransformer::Creating new Transformer Using Transformer Factory");
            Transformer t = getTransformerFactory().newTransformer();
            log.debug("XSLT Transformer::createTransformer::transformer created" + ( (t != null) ? "transformer Is not null" : "transformer is NULL"));
            log.debug("XSLT Transformer::createTransformer::Transformer Crated");
            return t;
        }
        else {
        	log.debug("XSLT Transformer::createTransformer::Source is NOT NULL creating transformer using templates");
        	Transformer t =  getTemplates().newTransformer();
        	log.debug("XSLT Transformer::createTransformer::transformer created" + ( (t != null) ? "transformer Is not null" : "transformer is NULL"));
        	return t;
        }
    }

    /**
     * A hook to allow the transformer to be configured from the current
     * exchange and inbound message
     */
    protected void configureTransformer(Transformer transformer, MessageExchange exchange, NormalizedMessage in) {
        for (Iterator iter = exchange.getPropertyNames().iterator(); iter.hasNext();) {
            String name = (String) iter.next();
            Object value = exchange.getProperty(name);
            transformer.setParameter(name, value);
        }
        for (Iterator iter = in.getPropertyNames().iterator(); iter.hasNext();) {
            String name = (String) iter.next();
            Object value = in.getProperty(name);
            transformer.setParameter(name, value);
        }
        if (xsltParameters != null) {
            for (Iterator iter = xsltParameters.keySet().iterator(); iter.hasNext();) {
                String name = (String) iter.next();
                Object value = xsltParameters.get(name);
                transformer.setParameter(name, value);
            }
        }
        transformer.setParameter("exchange", exchange);
        transformer.setParameter("in", in);
        transformer.setParameter("component", this);
    }

}

